/*
	DAPA.C

	Direct AVR Parallel Access (c) 1999
	
	Originally written by Sergey Larin.
	Corrected by:
	  Denis Chertykov, 
	  Uros Platise and (uros.platise@ijs.si)
	  Marek Michalkiewicz
	  Davor Jadrijevic
          Alexander Popov (sasho@vip.orbitel.bg)

  CHANGES:
        added the dirty hack, 07.02.00 robertle
*/

#ifndef NO_DAPA
//#define DEBUG

#include <stdio.h>
#include <stdlib.h>
#include <linux/serial_reg.h>

/* sys/io.h is the correct file.  If your system doesn't have that
   file, something is broken.
*/
/*
#ifdef __BROKEN_LINUX
#include <asm/io.h>
//#include <linux/lp.h>
#else
#include <sys/io.h>
//#include <linux/parport.h>
#endif
*/
#include "io_uisp.h"

#include <unistd.h>
#ifdef DIRTYHACK
#define __KERNEL__
#endif
#include <linux/lp.h>
#ifdef DIRTYHACK
#undef __KERNEL__
#endif
#include <signal.h>
#include <sys/time.h>
#include "Global.h"
#include "Error.h"
#include "DAPA.h"

/* Parallel Port Base Address
*/
#define BASE        parport_base

/* Alex's Direct Avr Parallel Access 
*/
#define DAPA_SCK    LP_PSTROBE	/* base + 2 */
#define DAPA_RESET  LP_PINITP	/* base + 2 */
#define DAPA_DIN    LP_PBUSY	/* base + 1 */
#define DAPA_DOUT   0x1		/* base */

/* Davor's Direct AVR Serial Access 
*/
#define DASA_MOSI   UART_LCR_SBC  /* TxD (0x40, base+3) MOSI */
#define DASA_RESET  UART_MCR_RTS  /* RTS (0x02, base+4) RESET */
#define DASA_SCK    UART_MCR_DTR  /* DTR (0x01, base+4) SCK & POWER */
#define DASA_MISO   UART_MSR_CTS  /* CTS (0x10, base+6) MISO */

/* STK200 Direct Parallel Access 
*/
#define STK2_ENA1   0x04    	/* D2 (base) - enable for RESET, MISO */
#define STK2_ENA2   0x08    	/* D3 (base) - enable for SCK, MOSI */
#define STK2_SCK    0x10    	/* D4 (base) */
#define STK2_DOUT   0x20    	/* D5 (base) */
#define STK2_RESET  0x80    	/* D7 (base) */
#define STK2_DIN    0x40    	/* ACK (base + 1) */

/* Altera Byte Blaster Port Configuration
*/
#define ABB_EN      LP_PAUTOLF	/* low active */
#define ABB_LPAD    0x80	/* D7: loop back throught enable auto-detect */
#define ABB_SCK	    0x01	/* D0: TCK */
#define ABB_RESET   0x02	/* D1: TMS */
#define ABB_DOUT    0x40	/* D6: TDI */
#define ABB_DIN	    LP_PBUSY	/* BUSY: TDO */

/* Atmel AVR ISP
*/
#define AVRISP_ENA     0x40                     /* D6 (base) - enable for MISO, MOSI, SCK */
#define AVRISP_SCK     0x10                     /* D4 (base) */
#define AVRISP_DOUT    0x20                     /* D5 (base) */
#define AVRISP_RESET   0x80                     /* D7 (base) */
#define AVRISP_DIN     0x40       		/* ACK (pin 10) (base + 1) */
#define AVRISP_DNG_IN  LP_POUTPA  		/* PO  (pin 12) (base + 1) - dongle input */
#define AVRISP_DNG_OUT 0x08                     /* D3 (base) - dongle output */
#define AVRISP_DNG_IN1 LP_PBUSY      		/* BUSY (pin 11) (base + 1) */


static volatile int yes_alarm;
static void alarm( int k ){
  yes_alarm = 0;
}

void TDAPA::DelayFunc(){
  volatile int i;
  for (i=0;i<100;i++);
}

void TDAPA::BigDelay(){
  static volatile int n;
  n = n_per_ms/25000 + 1;
  while (n){DelayFunc();n--;}
}

long TDAPA::CalcDelay(){ // pause x ms
  long n=0;
  struct itimerval it;
  it.it_interval.tv_sec = 0;
  it.it_value.tv_sec = 0;
  it.it_interval.tv_usec = 0;
  it.it_value.tv_usec = 100000;
  signal(SIGALRM, alarm);
  setitimer(ITIMER_REAL, &it, NULL);
  yes_alarm = 1;
  while (yes_alarm){
    DelayFunc(); n++;
  }
  return n;
}

void TDAPA::Reset(){
  switch(pa_type){
  case PAT_DAPA:
    outb(DAPA_SCK, BASE+2); usleep(1000);      /* sck = 0, reset = 0 */
    outb(DAPA_SCK | DAPA_RESET, BASE+2);       /* sck = 0, reset = 1 */
    usleep(1000);
    outb(DAPA_SCK, BASE+2); usleep(1000);      /* sck = 0, reset = 0 */
    break;

  case PAT_DASA:
    outb(UART_LCR_WLEN8 | UART_LCR_SBC, UART_LCR+BASE); /* select IER and FCR */
    outb(UART_MCR_DTR | UART_MCR_RTS, UART_MCR+BASE); /* enable port, no int. */
    outb(0, UART_IER+BASE); /* disable int */
    outb(UART_FCR_CLEAR_RCVR | UART_FCR_CLEAR_XMIT, UART_FCR+BASE); /* disable fifo */
    inb(UART_LSR+BASE); /* read int flags to clear them for good */
    inb(UART_RX+BASE); /* read int flags to clear them for good */
    inb(UART_IIR+BASE); /* read int flags to clear them for good */
    inb(UART_MSR+BASE); /* read int flags to clear them for good */
    usleep(100000); /* clk = sck = reset = 1 */
    outb(0, UART_LCR+BASE); usleep(1000); /* clk = 0, sck = 1, reset = 1 */
    outb(DASA_RESET, UART_MCR+BASE); usleep(1000); /* clk = 0, sck = 0, reset = 1 */
    outb(0, UART_MCR+BASE); usleep(1000); /* clk = 0, sck = 0, reset = 0 */
    break;

  case PAT_STK200:
    outb(STK2_ENA1 | STK2_ENA2, BASE);    	/* leave MCU */
    usleep(30000);
    outb(STK2_ENA2, BASE); usleep(1000);	/* sck = Z, reset = 0 */
    outb(0, BASE); usleep(1000);  		/* sck = 0, reset = 0 */
    outb(STK2_RESET, BASE); usleep(1000); 	/* sck = 0, reset = 1 */
    outb(0, BASE); usleep(1000);         	/* sck = 0, reset = 0 */       
    break;
    
  case PAT_ABB:
    outb(0,BASE+2); usleep(30000);		/* leave MCU */
    outb(0,BASE); 				/* reset=sck=0 */
    outb(ABB_EN,BASE+2); usleep(30000);		/* enable 244 driver */
    outb(ABB_RESET,BASE); usleep(30000);	/* reset=1 */
    outb(0,BASE); usleep(30000);		/* reset=0*/    
//    fprintf(stderr, "quiting ... test\n"); exit(1);
    break;

  case PAT_AVRISP:
    outb(AVRISP_ENA, BASE);    			/* leave MCU */
    usleep(30000);
    outb(0, BASE); usleep(1000);  		/* sck = 0, reset = 0 */
    outb(AVRISP_RESET, BASE); usleep(1000); 	/* sck = 0, reset = 1 */
    outb(0, BASE); usleep(1000);         	/* sck = 0, reset = 0 */       
    break;

  }

  usleep(20000);
}

void TDAPA::JustReset(){
  switch(pa_type){
  case PAT_DAPA:
    outb(DAPA_SCK | DAPA_RESET, BASE+2);       /* sck = 0, reset = 1 */
    break;
    
  case PAT_DASA:
    Info(0, "DASA: Unimplemented function.\n");
    break;    

  case PAT_STK200:
    outb(STK2_ENA2 | STK2_RESET, BASE);                /* sck = Z, reset = 1 */
    break;
    
  case PAT_ABB:
    outb(ABB_EN,BASE+2);                       /* enable 244 driver */
    outb(ABB_RESET,BASE);                      /* reset=1 */
    break;

  case PAT_AVRISP:
    outb(AVRISP_ENA | AVRISP_RESET, BASE);     /* Leave MCU, reset = 1 */
    break;
  }
  usleep(20000);
}

void TDAPA::SCKtoRetryEnable(){
  switch(pa_type){
  case PAT_DAPA:
    outb(0, BASE+2); BigDelay();
    outb(DAPA_SCK, BASE+2); BigDelay();
    break;

  case PAT_DASA:
    outb (DASA_SCK, UART_MCR+BASE); BigDelay ();  /* sck = 1 */
    outb (0, UART_MCR+BASE); BigDelay ();  /* sck = 0 */
    break;

  case PAT_STK200:
    outb(STK2_SCK, BASE); BigDelay();  	/* sck = 1 */
    outb(0, BASE); BigDelay();  	/* sck = 0 */
    break;

  case PAT_ABB:
    outb(ABB_SCK, BASE); BigDelay();
    outb(0, BASE); BigDelay();
    break;    

  case PAT_AVRISP:
    outb(AVRISP_SCK, BASE); BigDelay(); /* sck = 1 */
    outb(0, BASE); BigDelay();		/* sck = 0 */
    break;    
  }
}

unsigned char TDAPA::SendRecv(unsigned char b){
  int i, mask, received=0;
  
  for (i=0, mask = 0x80; i<8; i++, mask >>= 1){
     switch(pa_type){
     case PAT_DAPA:
       outb(b & mask ? DAPA_DOUT : 0, BASE); BigDelay();
       outb(0, BASE+2); BigDelay ();
       if (!(inb (BASE+1) & DAPA_DIN)){received |= mask;}
       outb(DAPA_SCK, BASE+2); BigDelay();
       break;
     
     case PAT_DASA:
       outb((b & mask) ? DASA_MOSI : 0, UART_LCR+BASE); /* send data bit, sck = 0 */
       outb(DASA_SCK, UART_MCR+BASE); /* send data bit, sck = 1 */
       if(longsck > -2)
        BigDelay();
       if(i == 0 && longsck > 5) usleep(longsck); /* allow recharging of the capacitor */
       if (inb(UART_MSR+BASE) & DASA_MISO){received |= mask;} /* read data bit */  
       outb(0, UART_MCR+BASE); /* send data bit, sck = 0 */
       outb(DASA_MOSI, UART_LCR+BASE); /* power it up */
       if(longsck > -1)
        BigDelay();
       break;

     case PAT_STK200:
       outb((b & mask) ? STK2_DOUT : 0, BASE);  /* send data bit, sck = 0 */
       BigDelay();
       if (inb(BASE + 1) & STK2_DIN){received |= mask;} /* read data bit */
       						/* send data bit, sck = 1 */
       outb(STK2_SCK | ((b & mask) ? STK2_DOUT : 0), BASE);  
       BigDelay();
       outb((b & mask) ? STK2_DOUT : 0, BASE);	/* send data bit, sck = 0 */
       BigDelay();
       break;
       
     case PAT_ABB:
       outb((b & mask) ? ABB_DOUT : 0, BASE);  /* send data bit, sck = 0 */
       BigDelay();
       if (!(inb(BASE + 1) & ABB_DIN)){received |= mask;} /* read data bit */
       						/* send data bit, sck = 1 */
       outb(ABB_SCK | ((b & mask) ? ABB_DOUT : 0), BASE);  
       BigDelay();
       outb((b & mask) ? ABB_DOUT : 0, BASE);	/* send data bit, sck = 0 */
       BigDelay();
       break;

     case PAT_AVRISP:
       outb((b & mask) ? AVRISP_DOUT : 0, BASE);  /* send data bit, sck = 0 */
       BigDelay();
       if (inb(BASE + 1) & AVRISP_DIN){received |= mask;} /* read data bit */
       						/* send data bit, sck = 1 */
       
       outb(AVRISP_SCK | ((b & mask) ? AVRISP_DOUT : 0), BASE);  
       BigDelay();
       outb((b & mask) ? AVRISP_DOUT : 0, BASE);	/* send data bit, sck = 0 */
       BigDelay();
       break;
     }
  
  }
  return received;
}

int TDAPA::Send (unsigned char* queue, int queueSize, int rec_queueSize=-1){
  unsigned char *p = queue, ch;
  int i = queueSize;
  
  if (rec_queueSize==-1){rec_queueSize = queueSize;}
#ifdef DEBUG
  printf ("send(recv): ");
#endif
  while (i--){
#ifdef DEBUG
    printf ("%02X(", (unsigned int)*p);
#endif    
    ch = SendRecv(*p);
#ifdef DEBUG    
    printf ("%02X) ", (unsigned int)ch);
#endif    
    *p++ = ch;
  }
#ifdef DEBUG  
  printf ("\n");
#endif  
  return queueSize;
}

TDAPA::TDAPA(): 
  parport_base(0x378){
    
  const char* val;  
  /* Parse Command Line Switches */
  longsck = 0;
  if((val=GetCmdParam("-dlongsck"))!=NULL)
  {
   longsck = strtoul(val, NULL, 0);
  }
  if ((val=GetCmdParam("-dlpt"))!=NULL){
    if (!strcmp(val, "1")){parport_base = 0x278;}
    else if (!strcmp(val, "2")){parport_base = 0x378;}
    else if (!strcmp(val, "3")){parport_base = 0x3bc;}    
    else{
      parport_base = strtol(val, NULL, 0);
      if (parport_base!=0x278 && parport_base!=0x378 && parport_base!=0x3bc){
        throw Error_Device("Bad device address.");
      }
    }
  }

  /* Parse Command Line Switches */
  if ((val=GetCmdParam("-dcom"))!=NULL){
    if (!strcmp(val, "1")){parport_base = 0x3f8;}
    else if (!strcmp(val, "2")){parport_base = 0x2f8;}
    else if (!strcmp(val, "3")){parport_base = 0x3e8;}    
    else if (!strcmp(val, "4")){parport_base = 0x2e8;}    
    else{
      parport_base = strtol(val, NULL, 0);
      if (parport_base!=0x3f8 && parport_base!=0x2f8 &&
          parport_base!=0x3e8 && parport_base!=0x2e8){
        throw Error_Device("Bad device address.");
      }
    }
  }

  
  /* Enable Parallel Port */
  if (GetCmdParam("-dapa", false)){pa_type = PAT_DAPA;}
  else if (GetCmdParam("-dasa", false)){pa_type = PAT_DASA;}
  else if (GetCmdParam("-dstk200", false)){pa_type = PAT_STK200;}
  else if (GetCmdParam("-dabb", false)){pa_type = PAT_ABB;}
  else if (GetCmdParam("-davrisp", false)){pa_type = PAT_AVRISP;}
  else{
    throw Error_Device("Direct Parallel Access not defined.");
  }

  n_per_ms = CalcDelay();
  Info(3, "Number of delay loops per 100 millisecond: %ld\n", n_per_ms);
  
  if (::ioperm(BASE, 8, 1) == 0){  
  
    if (!GetCmdParam("--reset",false)) Reset();
    
    /* we don't need root permissions anymore, so we drop them for
       security reasons. M.Matuszek, 30.01.2000 
    */
    Info(3, "%s\n", 
            "port access granted, dropping permissions...");
    setgid (getgid ());  /* reset group_id to that of calling process */
    setuid (getuid ());  /* reset user_id to that of calling process */
  } else{
    perror("ioperm()"); throw Error_Device("DPA","Cannot open port.");
  }

  if (pa_type == PAT_AVRISP) { /* Check if dongle is present */
    unsigned char avrisp_present=1;

    outb(AVRISP_DNG_OUT | AVRISP_ENA, BASE);
    if (!(inb(BASE+1)&AVRISP_DNG_IN)) avrisp_present=0; 
    outb(AVRISP_ENA, BASE);
    if (inb(BASE+1)&AVRISP_DNG_IN) avrisp_present=0;
    if (!(inb(BASE+1)&AVRISP_DNG_IN1)) avrisp_present=0;
    if (!avrisp_present) {
       throw Error_Device("AVR ISP is not present.");
    }
  }
}

TDAPA::~TDAPA(){
  switch(pa_type){  
  case PAT_DAPA:   outb(DAPA_SCK | DAPA_RESET, BASE+2); break;
  case PAT_DASA:   outb(DASA_RESET | DASA_SCK, UART_MCR+BASE);
                   outb(DASA_MOSI, UART_LCR+BASE); break;
  case PAT_STK200: outb(STK2_ENA1 | STK2_ENA2 | STK2_RESET, BASE); break;
  case PAT_ABB:    outb(0, BASE+2); outb(ABB_RESET, BASE); break;
  case PAT_AVRISP: outb(AVRISP_ENA | AVRISP_RESET, BASE);   break;
  }
}

#endif
/* eof */
